@@ -7,8 +7,8 @@ mod trapcode;
 pub mod traphandlers;
 
 pub use trapcode::TrapCode;
+pub use traphandlers::resume_panic;
 pub use traphandlers::{
     catch_traps, catch_traps_with_result, raise_lib_trap, raise_user_trap, wasmer_call_trampoline,
-    TlsRestore, Trap, TrapHandler, TrapHandlerFn,
+    TlsRestore, Trap,
 };
-pub use traphandlers::{init_traps, resume_panic};
